;=======================================================================================
;                              Gyro Controller 
;=======================================================================================
;                              by Bruce Abbott 
;
; for Microchip PIC12F615 or PIC12F683
;
; Features:
;
;     - Gain adjustment 
;     - Heading Hold on/off  
;
; Changes:
;
; 2010-07-10 V0.0 created
;
		ifdef	__12F615
                PROCESSOR PIC12F615
                INCLUDE   <P12F615.inc>
CMCON		equ	CMCON0
VREN		equ	CMVREN
		endif

		ifdef	__12F683
                PROCESSOR PIC12F683
                INCLUDE   <P12F683.inc>
		endif

                radix     dec

		errorlevel -302,-305	; not reporting bank<>0, dest=file

;#DEFINE SIMULATE	; for debugging only!


#define VERSION " G Y R O C O N   0 . 0"

; =====================================================================================
; Configuration is:
;   Master Clear pin is disabled (used for servo pulse input)
;   Code Protection is OFF
;   Watchdog Timer is ON
;   Oscillator is Internal RC
;
		ifdef	__12F615
		__config  _MCLRE_OFF & _INTRC_OSC_NOCLKOUT & _IOSCFS_8MHZ
		endif

		ifdef	__12F683
		__config  _MCLRE_OFF & _INTRC_OSC_NOCLKOUT
		endif

; GPIO register bits 
potbit          EQU       GP0 ; pin 7. gain potentiometer 
switchbit       EQU       GP1 ; pin 6. Heading Hold switch
outbit	        EQU       GP2 ; pin 5. Gyro Control output pulse
inbit           EQU       GP3 ; pin 4. Gyro input pulse
;               EQU       GP4 ; pin 3.  
;               EQU       GP5 ; pin 2. 

; I/O direction register value
TrisBits        EQU       (1<<potbit)|(1<<switchbit)|(1<<inbit)	; inputs

; Bits to be set with the OPTION instruction
;     7 = Weak pullups disabled
;     6 = Interrupt on Rising Edge
;     5 = Timer 0 source internal
;     4 = Which edge is don't care
;   3~1 = Prescaler to watchdog, set to give 16*17 = 272ms timeout

OptionBits      EQU       B'11011100'

; =====================================================================================
;                        General Purpose Register Definitions  
; =====================================================================================
		ifdef	__12F615
		CBLOCK	0x40
		ENDC
		endif

		ifdef	__12F683
		CBLOCK	0x20
		ENDC
		endif

		CBLOCK
                flags           ; Various boolean flags
		pot:2		; potentiometer position (16 bits)
		samples		; number of samples per pot measurement
		gain		; gyro gain 0-255
		pulsewidth	; output pulse width 0-255 = 1~2mS
                temp		; general purpose variable
		ENDC

; Bits in Flags

SyncBit 	EQU       1	; gyro input pulse detected
HHonBit		EQU	  2	; Heading Hold switched on

;==================================================================================
;                     Macro for generating short time delays
;===================================================================================
NO_OP           MACRO   count
NO_OP_COUNT     SET     count
                WHILE   NO_OP_COUNT>1
		goto	$+1		; 2 clocks
NO_OP_COUNT     SET     NO_OP_COUNT-2
                ENDW
		IF	NO_OP_COUNT
		nop			; 1 clock
		ENDIF
                ENDM

;==================================================================================
;                       16 Bit Math Macros  (little-endian)
;==================================================================================

; 16 bit unsigned addition with carry out.
;
; Operation: DST = DST + SRC                       
;
; DST is replaced, SRC is preserved
; Carry Set if SRC + DST > 65535
;
ADD16   MACRO   SRC,DST 
        MOVF    (SRC),W         ; Get low byte
        ADDWF   (DST),F         ; Add to destination
        MOVF    (SRC)+1,W       ; Get high byte
        BTFSC   STATUS,C        ; Check for carry
        INCF    (SRC)+1,W       ; Add one for carry
        ADDWF   (DST)+1,F       ; Add high byte into DST
        ENDM

; 16 bit unsigned subtraction with carry out.
; Word format is little endian (LSB at lower address)
;
; Operation: DST = DST - SRC
;
; DST is replaced, SRC is preserved
; Carry Set if DEST >= SRC (result >=0) 
;
SUB16   MACRO   SRC,DST
        MOVF    (SRC),W         ; Get low byte of subtrahend
        SUBWF   (DST),F         ; Subtract DST(low) - SRC(low)
        MOVF    (SRC)+1,W       ; Get high byte of subtrahend
        BTFSS   STATUS,C        ; If there was a borrow then
        INCF    (SRC)+1,W       ; decrement high byte of dst
        SUBWF   (DST)+1,F       ; Subtract the high byte 
        ENDM

; 16 bit copy
;
MOV16   MACRO   SRC,DST
        MOVF    SRC,W
        MOVWF   DST
        MOVF    SRC+1,W
        MOVWF   DST+1
        ENDM

; 16 bit clear 
;
CLR16   MACRO   VAR16
        CLRF    VAR16
        CLRF    VAR16+1
        ENDM
        
; 16 bit Logical Shift Right (= divide by 2)
;
LSR16   MACRO   VAR16
        CLRC		        ; Clear carry
        RRF     (VAR16)+1       ; Rotate high byte right
        RRF     (VAR16)         ; Rotate low byte right
        ENDM
        
; 16 bit Logical Shift Left  (= multiply by 2)
;
LSL16   MACRO   VAR16
        CLRC		        ; Clear carry
        RLF     (VAR16)         ; Rotate low byte left
        RLF     (VAR16)+1       ; Rotate upper byte left
        ENDM  

; 16 bit Rotate Left
;
RL16	MACRO	VAR16
	RLF	(VAR16)
	RLF	(VAR16)+1
	ENDM

; 16 bit Rotate Right
;
RR16	MACRO	VAR16
	RRF	(VAR16)+1
	RRF	(VAR16)
	ENDM

;==================================================================================
;                                     RESET
;==================================================================================
;
Coldstart:      ORG     0                

		goto	start

		org	0x8
                dw      VERSION



;==================================================================================
;                         Wait for end of next PPM pulse
;==================================================================================
;
waitpulse:	clrwdt 
		btfsc   GPIO, inbit	; wait for end of last pulse
                goto    waitpulse
waitpulse1: 	clrwdt
		btfss   GPIO, inbit	; wait for start of next pulse
                goto    waitpulse1
waitpulse2:	clrwdt 
		btfsc   GPIO, inbit	; wait for end of next pulse
                goto    waitpulse2
		return

;=================================================================================
;                Measure voltage on an A/D input 
;=================================================================================
; input: w = pointer to volts variable  
;
; output: pot = voltage (0~1023)  
;  
GetVolts:	movwf	FSR		; FSR = pot
		movlw	(200-3)/4
v_delay:	addlw	-1
		skpnc			; wait 100uS for A/D input to stabalize
		goto	v_delay	
		movlw	8		; 8 x oversampling
		movwf	samples
		clrf	INDF		; volts (low byte) = 0		  
		incf	FSR		
		clrf	INDF		; volts (high byte) = 0
v_read:
		bsf	ADCON0,GO	; start A/D conversion
v_wait:
		btfsc	ADCON0,GO	; wait until conversion complete
		goto	v_wait
		movf	ADRESH,w	; A/D result (high byte)
		addwf	INDF		; add to volts (high byte)
		banksel	ADRESL
		movf	ADRESL,w	; A/D result (low byte)
		banksel	0
		decf	FSR
		addwf	INDF		; add to pot (low byte)
		incf	FSR
		skpnc			; low byte overflow?
		incf	INDF		; yes, add 1 to high byte
		decfsz	samples
		goto	v_read		; next sample 
		movlw	3
v_div:		clrc
		rrf	INDF
		decf	FSR
		rrf	INDF		;  shift right * 3 = divide by 8
		incf	FSR
		addlw	-1
		skpz
		goto	v_div
		return

;=================================================================================
;                         Get Gain Potentiometer position
;=================================================================================
;
GetGain:	bcf	ADCON0,CHS1
		bcf	ADCON0,CHS0	; select A/D channel 0
		movlw	pot
		call	GetVolts	; get pot voltage (10 bit)
		RR16	pot
		RR16	pot		; rotate right * 2 = / 4 = convert to 8 bit
		movf	pot,w
		movwf	gain		; gyro gain = pot position (0~255)
		return
 
;================================================================================
;                         Generate gyro control output pulse  
;================================================================================
; input: pulsewidth 0~255 = output pulse width 1110uS~1890uS 
;
DoPulse:	bsf	GPIO,outbit
		movlw	(1120/10)-1
dp_1:		NO_OP	16
		addlw	-1
		skpnc			; delay 1120uS
		goto	dp_1
		movf	pulsewidth,w
dp_2:		NO_OP	2
		addlw	-1
		skpnc			; delay * pulsewidth (3uS resolution)
		goto	dp_2
		bcf	GPIO,outbit	; finish output pulse
                return              

;------------------------------------------------------------------------------------
;                           END OF SUBROUTINES
;------------------------------------------------------------------------------------


                ORG	256               

; Initialise GPIO register

start:          clrf    GPIO		; all outputs will be low
                movlw   TrisBits
                tris    GPIO		; set I/O directions

 		banksel	ANSEL		; Register bank 1

		movlw	((1<<ADCS2)|(1<<ADCS0)|(1<<ANS0))
		movwf	ANSEL		; clk/8, Analog I/O on GP0 

		ifdef	__12F683
		bsf	OSCCON,IRCF0	; switch to 8MHz clock
		endif

		banksel	0		; register bank 0

;		movlw	1<<CMON|1<<CMR  ; Comparator enabled, in- to GP1, in+ to VREF       

;
; Move the prescaler from tmr0 to the watchdog, without accidental resets.
;
                clrwdt                    
                clrf      TMR0            
                movlw     OptionBits | 7  
                option                    
                clrwdt
                movlw     OptionBits
                option
                clrwdt

; Set up Comparator
		movlw	b'00000111'		; comparator off
		movwf	CMCON0

; set up A/D convertor
		movlw	(1<<ADFM)|(1<<ADON) ; right-justify, A/D on
		movwf	ADCON0

; ready to go!            
		clrf	flags

mainloop:	call	GetGain		; get gain setting
		clrc
		rrf	gain		; gain = 0~127
		movlw	128		; pulse center = 1.5mS 
		movwf	pulsewidth	
		movf	gain,w
		btfsc	GPIO,switchbit	; in heading hold mode?
		goto	headinghold	; yes

		subwf	pulsewidth	; pulse = 127~0 (rate mode)
		goto	sync

headinghold:	addwf	pulsewidth	; pulse = 128~255 (heading hold mode)

sync:		call	waitpulse	; wait for end of gyro input pulse
		call	DoPulse		; create control pulse
		goto	mainloop

                END

